class TrieNode:
    """
    This class implements a node in a trie. It's usage outside of a trie is not recommended although no one is
    stopping you.

    NOTE: This supports ASCII character set only.

    Structurally, it contains a fixed size list internally, initialized to all Nones. Whenever a character in inserted,
    it creates another TrieNode object at the index corresponding to the ascii character.
    """
    ASCII_CHARSET_SIZE = 128

    def __init__(self):
        """
        Constructor. Needs no parameters
        """
        self.storage = [None] * TrieNode.ASCII_CHARSET_SIZE
        self.end_marker = False

    def __contains__(self, char: str) -> bool:
        """
        Implements the contains functionality. You can do `'a' in trie_node_object`.
        :param char: The character to test
        :return: Boolean
        """
        if len(char) > 1:
            raise ValueError("'{}' is not a single character.".format(char))
        return self.storage[ord(char)] is not None

    def __getitem__(self, char: str) -> 'TrieNode':
        """
        This implements the index into the object. It returns the trie node object in the index corresponding to
        the ASCII value of the character. `None` would indicate absence of that character in the node object.
        :param char:
        :return:
        """
        if len(char) > 1:
            raise ValueError("'{}' is not a single character.".format(char))
        return self.storage[ord(char)]

    def __iter__(self):
        """
        Implements the iter functionality. It iterates over the internal storage.
        :return: An iterator.
        """
        return iter(self.storage)

    def set_char(self, char: str, end_marker: bool = False):
        """
        This sets a character inside a trie. If the character already existed at that position, a end marker is installed
        if needed.
        :param char:
        :param end_marker:
        :return:
        """
        if len(char) > 1:
            raise ValueError("'{}' is not a single character.".format(char))

        if char not in self:
            self.storage[ord(char)] = TrieNode()
        node = self.storage[ord(char)]
        node.end_marker = end_marker or node.end_marker


class Trie:
    """
    This class is the implement of a Trie data structure, for the ASCII character set. The contents are case sensitive.
    """

    def __init__(self):
        """
        Constructor. This takes no arguments
        """
        self.root = TrieNode()

    @staticmethod
    def _insert_callback(word: str, start_pos: int, node: TrieNode):
        if start_pos == len(word) - 1:
            node.set_char(word[start_pos], True)
        else:
            node.set_char(word[start_pos])
            Trie._insert_callback(word, start_pos + 1, node[word[start_pos]])

    @staticmethod
    def _search_callback(word: str, pos: int, node: TrieNode) -> bool:
        if pos == len(word):
            return node
        else:
            return Trie._search_callback(word, pos + 1, node[word[pos]]) if (word[pos] in node) else None

    @staticmethod
    def _collect_branches(node: TrieNode) -> list:
        if node is None:
            return ['']
        sub_strings = []
        valid_next_nodes = filter(lambda elem: elem[1] is not None, enumerate(node))
        for enum, nxt in valid_next_nodes:
            char = chr(enum)
            end_marker = nxt.end_marker
            if end_marker:
                sub_strings.append(char)

            more_substrings = ["{}{}".format(char, s) for s in Trie._collect_branches(nxt)]
            sub_strings.extend(more_substrings)
        return sub_strings

    def insert(self, word: str):
        """
        This function inserts a word into the trie
        :param word: The word to insert into the trie
        :return
        """
        Trie._insert_callback(word, 0, self.root)

    def __contains__(self, word: str) -> bool:
        """
        This implements the contains method. Use it like "word" in my_trie_struct
        :param word:
        :return: `boolean` whether the trie contains the word
        """
        terminal_node = Trie._search_callback(word, 0, self.root)
        return False if terminal_node is None else terminal_node.end_marker

    def __getitem__(self, word: str) -> list:
        """
        This returns the tree under the passed index. This is useful for implementing type-ahead.
        :param word: The index
        :return: A list of words under the index
        """
        terminal_node = Trie._search_callback(word, 0, self.root)
        entries = []
        if terminal_node.end_marker:
            entries.append(word)
        entries.extend(["{}{}".format(word, s) for s in Trie._collect_branches(terminal_node)])
        return entries



from inspect import getfullargspec


def flatten(item):
    """
    Recursively flattens a nested set of collections. Example
    (1, (2, 3)) would be flattened to [1,2,3]

    Dictionaries are flattened to tuples. Example
    {
            '213': {
                'abc': 'def',
                'xyz': 'ccc'
            },
            '214': {
                'abc': 'xyz',
                'papaya': [1, (2, 3)]
            }
        }
    would be flattened to
    [
        ('213', 'abc', 'def'),
        ('213', 'xyz', 'ccc'),
        ('214', 'abc', 'xyz'),
        ('214', 'papaya', 1),
        ('214', 'papaya', 2),
        ('214', 'papaya', 3)
    ]
    :param item: A possibly nested iterable
    :return: A flat list
    """
    if type(item) == str:
        return [item]
    elif type(item) == dict:
        flattened = []
        for k, v in item.items():
            flattened_values = flatten(v)
            for flattened_value in flattened_values:
                if type(flattened_value) != str and hasattr(flattened_value, '__iter__'):
                    flattened.append(
                        tuple(
                            [k] + [vv for vv in flattened_value]
                        )
                    )
                else:
                    flattened.append((k, flattened_value))
        return flattened
    elif hasattr(item, "__iter__"):
        flattened = []
        for itm in item:
            flattened.extend(flatten(itm))
        return flattened
    else:
        return [item]


def flatten_to_tuple(item):
    """
    Recursively flattens a nested set of collections, except tuples. Example
    (1, (2, 3)) would be flattened to [1,(2,3)]. However [[1, 2], [(3,4), 5] will
    be flattened to [1, 2, (3, 4), 5]

    Dictionaries are flattened to tuples. Example
    {
            '213': {
                'abc': 'def',
                'xyz': 'ccc'
            },
            '214': {
                'abc': 'xyz',
                'papaya': [1, (2, 3)]
            }
        }
    would be flattened to
    [
        ('213', ('abc', 'def')),
        ('213', ('xyz', 'ccc')),
        ('214', ('abc', 'xyz')),
        ('214', ('papaya', 1)),
        ('214', ('papaya', (2, 3)))
    ]

    :param item: A possibly nested iterable
    :return: A flat list
    """
    if type(item) == str:
        return [item]
    elif type(item) == tuple:
        return [item]
    elif type(item) == dict:
        flattened = []
        for k, v in item.items():
            flattened_values = flatten_to_tuple(v)
            for flattened_value in flattened_values:
                if type(flattened_value) != tuple and type(flattened_value) != str and hasattr(flattened_value, '__iter__'):
                    flattened.append(
                        tuple(
                            [k] + [vv for vv in flattened_value]
                        )
                    )
                else:
                    flattened.append((k, flattened_value))
        return flattened
    elif hasattr(item, "__iter__"):
        flattened = []
        for itm in item:
            flattened.extend(flatten_to_tuple(itm))
        return flattened
    else:
        return [item]


def _all_subsequences_from_idx(arr: list, start: int, memo: dict) -> list:
    if start in memo:
        return memo[start]
    start_elem = arr[start]
    all_subs = [[start_elem]]
    for it in range(start+1, len(arr)):
        sub_subsequences = _all_subsequences_from_idx(arr, it, memo)
        all_subs.extend(
            [[start_elem] + s for s in sub_subsequences]
        )
    memo[start] = all_subs
    return all_subs


def subsequence(arr: list) -> list:
    """
    This function returns all subsequences of a list. A subsequence is defined as any sublist of a list such that the
    elements in the subsequence are in the same order as the original list.
    Example: Subsequences of [1, 2, 3] are [1], [2], [3], [1, 2], [1, 3], [2, 3]

    :param arr: The list whose subsequences are needed.
    :return: A list of subsequences (list[list])
    """
    all_subs = []
    memo = dict()
    for it in range(len(arr)):
        all_subs.extend(_all_subsequences_from_idx(arr, it, memo))
    return all_subs


class _Foreach:
    def __init__(self, functor, iterable):
        num_args = len(getfullargspec(functor).args)
        if num_args != 1:
            raise TypeError(f"Passed function should take only 1 argument, instead it expects {num_args}.")
        if not hasattr(iterable, "__iter__"):
            raise TypeError(f"{iterable} is not iterable.")
        self._functors = [functor]
        self._iterable = iterable

    def then(self, functor):
        num_args = len(getfullargspec(functor).args)
        if num_args != 1:
            raise TypeError(f"Passed function should take only 1 argument, instead it expects {num_args}.")
        self._functors.append(functor)
        return self

    def __call__(self, *args, **kwargs):
        for item in self._iterable:
            temp = item
            for function in self._functors:
                temp = function(temp)


class _Transform:
    def __init__(self, functor, iterable):
        num_args = len(getfullargspec(functor).args)
        if num_args != 1:
            raise TypeError(f"Passed function should take only 1 argument, instead it expects {num_args}.")
        if not hasattr(iterable, "__iter__"):
            raise TypeError(f"{iterable} is not iterable.")
        self._functors = [functor]
        self._iterable = iterable
        self._internal_store = list()

    def then(self, functor):
        num_args = len(getfullargspec(functor).args)
        if num_args != 1:
            raise TypeError(f"Passed function should take only 1 argument, instead it expects {num_args}.")
        self._functors.append(functor)
        return self

    def __iter__(self):
        class _TransformIterator:
            def __init__(self, collection, functors: list):
                self._functors = functors
                self._it = iter(collection)

            def __next__(self):
                raw_item = self._it.__next__()
                transformed_item = raw_item
                for f in self._functors:
                    transformed_item = f(transformed_item)
                return transformed_item
        return _TransformIterator(self._iterable, self._functors)


def foreach(functor, iterable):
    """
    This function allows the user to apply a method on each element of an iterable. Unlike `map`, there is no side
    effect. So you cannot use it to transform your iterable into a new one.

    This is lazily evaluated, and allows chaining using `then` function. You need to call the `()` method to start
    evaluation.

    Example
    -------
    ```
    my_list = [1, 2, 3]
    foreach(lambda item: print(f"Parsing {item} in f1"), my_list)
        .then(lambda item: print(f"Parsing {item} in f2"))
        .then(lambda item: print(f"Parsing {item} in f3"))()

    Output will be
    Parsing 1 in f1
    Parsing 1 in f2
    Parsing 1 in f3
    Parsing 2 in f1 .....

    :param functor: The function to apply gto each element in an iterable.
    :param iterable: The iterable under processing.
    :return: A _ForEach object (this is internal). Just use the `.then` method to chain or just call it `()` to start
    execution.
    ```
    """
    return _Foreach(functor, iterable)


# TODO: Add filter capabilities
def transform(functor, iterable):
    """
    This function allows the user to apply a method on each element of an iterable, and produce a new one. The
    difference between this and Python's build in `map` function is that this allows chaining using `then` function.

    Example
    -------
    ```
    my_list = [1, 2, 3]
    transformed = transform(
            lambda item: item**2,
            my_list
        ).then(
            lambda item: item + 1
        ).then(
            lambda item: item * (-1)
        )
    transformed_list = list(transformed)
    print(transformed_list)

    Output will be
    [-2, -5, -10]

    :param functor: The function to apply gto each element in an iterable.
    :param iterable: The iterable under processing.
    :return: A _Transform object (this is internal). Just use the `.then` method to chain more transformer functions.
    You can pass the returned object to the constructor of any iterable just like you would with a `map` object.

    ```
    """
    return _Transform(functor, iterable)


def group_by(functor, iterable) -> dict:
    """
    This function groups an iterable into a dictionary ver much similar to Scala's groupBy method. This function will
    apply a callable to each item in the iterable to generate a key, and all the items that generate the same key will
    be grouped as values.

    Example
    -------
    ```
    my_list = [
        [1, "a"],
        [1, "b"],
        [2, "a"],
    ]
    grouped = group_by(
        lambda item: item[0],
        my_list
    )

    print(grouped)

    Output will be
    {
        1: [[1, "a"], [1, "b"]],
        2: [[2, "a"]]
    }

    :param functor: The function to apply gto each element in an iterable.
    :param iterable: The iterable under processing.
    :return: A _Transform object (this is internal). Just use the `.then` method to chain more transformer functions.
    You can pass the returned object to the constructor of any iterable just like you would with a `map` object.

    ```
    """
    retval = dict()
    for item in iterable:
        key = functor(item)
        if key not in retval:
            retval[key] = list()
        retval[key].append(item)
    return retval



class Result:
    """
    This class is supposed to be similar in principle to Rust's Result.
    """
    @classmethod
    def success(cls, value) -> 'Result':
        """
        This returns a success object with `val` set o `value`
        :param value:
        :return: A success `Result` object.
        """
        res = Result()
        res.is_success = True
        res.val = value
        return res

    @classmethod
    def err(cls, err_val) -> 'Result':
        """
        This returns a error object with `err` set o `err_val`
        :param err_val:
        :return: A failure `Result` object.
        """
        res = Result()
        res.is_success = False
        res.err = err_val
        return res

    def __call__(self, *args, **kwargs):
        if self.is_success:
            return self.val
        else:
            return self.err

    def __bool__(self):
        return self.is_success



import json
from abc import ABC, abstractmethod
from thesoup.utilityfunctions.collectionutils import flatten


class Edge:
    """
    This class is meant to represent an edge of a graph. It has 3 members: source, destination and a "property". This
    is left vague on purpose. Use the "property" to indicate weight or something else if you like.

    All the 3 members must define the `__str__` and `__eq__` methods.
    """
    def __init__(self, src, destination, ppt):
        self.src = src
        self.destination = destination
        self.ppt = ppt

    def __hash__(self) -> int:
        return int(
            hash(
                "{}.{}.{}".format(self.src, self.destination, self.ppt)
            )
        )

    def __eq__(self, other) -> bool:
        return type(other) == Edge and \
               self.src == other.src and \
               self.destination == other.destination and \
               other.ppt == self.ppt

    def __str__(self):
        return "{}.{}.{}".format(self.src, self.destination, self.ppt)

    def __gt__(self, other):
        if type(other) != Edge:
            raise TypeError("Cannot compare edge to non edge")
        return self.ppt > other.ppt


class Graph (ABC):
    """
    This is the interface for a directed graph
    """
    @abstractmethod
    def get_neighbours(self, item) -> list:
        """
        Returns the neighbours of a vertex
        :param item: Vertex whose neighbours to fetch
        :return:
        """
        pass

    @abstractmethod
    def __contains__(self, item):
        """
        Returns if a vertex or an edge exists in the graph.
        NOTE: To test the existence of an edge, pass an object of `Edge` type
        :param item: the vertex or edge to test
        :return:
        """
        pass

    @abstractmethod
    def vertices(self) -> set:
        """
        Returns the set of vertices
        """
        pass

    @abstractmethod
    def edges(self) -> set:
        """
        Returns the set of edges as `Edge` objects
        """
        pass


class MutableGraph (Graph):
    """
    This is the interfaces for a mutable directed graph. Aside from the basic interfaces for getting the neighbours and
    testing if the graph contains an element, it also provides interfaces to add vertices and edges to the graph
    """
    @abstractmethod
    def add_vertex(self, item):
        """
        Adds a vertex
        :param item: Vertex to add
        :return:
        """
        pass

    @abstractmethod
    def add_edge(self, edge: Edge):
        """
        The edge to add
        :param edge: An `Edge` struct
        :return:
        """
        pass


class AdjListDiGraph (MutableGraph):
    """
    Implementation of a directed graph using adjacency list
    """
    def __init__(self):
        self.storage = dict()

    """
    {}
    """.format(Graph.get_neighbours.__doc__)
    def get_neighbours(self, item) -> list:
        if item not in self.storage:
            return None
        else:
            return self.storage[item]

    """
    {}
    """.format(MutableGraph.add_vertex.__doc__)
    def add_vertex(self, item):
        if item not in self.storage:
            self.storage[item] = set()

    """
    {}
    """.format(MutableGraph.add_edge.__doc__)
    def add_edge(self, edge: Edge):
        if edge.src not in self.storage or edge.destination not in self.storage:
            raise ValueError("Either the source ({}) or destination ({}) of the edge is not in the graph".format(
                edge.src,
                edge.destination
            ))
        self.storage[edge.src].add((edge.destination, edge.ppt))

    def __contains__(self, item):
        """
        {} 
        """.format(Graph.__contains__.__doc__)
        if type(item) == Edge:
            return item.src in self.storage and (item.destination, item.ppt) in item.storage[item.src]
        else:
            return item in self.storage

    def vertices(self) -> list:
        """
        {}
        """.format(Graph.vertices.__doc__)
        return set(self.storage.keys())

    def edges(self) -> set:
        """
        {}
        """.format(Graph.edges.__doc__)
        return set(flatten([[Edge(v, e[0], e[1]) for e in edges] for v, edges in self.storage.items()]))

    @staticmethod
    def from_json(json_str: str):
        data = json.loads(json_str)
        graph = AdjListDiGraph()
        for v in data.keys():
            graph.add_vertex(v)

        for v, edges in data.items():
            for destination, ppt in edges:
                graph.add_edge(Edge(v, destination, ppt))

        return graph


class AdjListUndirectedDiGraph (AdjListDiGraph):
    """
    Implementation of a directed graph using adjacency list.
    NOTE: For this, the vertices have to have the __gt__ operator implemented.
    NOTE: This does allow multiple edges with different weights to exist between nodes.
    """

    def add_edge(self, edge: Edge):
        if edge.src not in self.storage or edge.destination not in self.storage:
            raise ValueError("Either the source ({}) or destination ({}) of the edge is not in the graph".format(
                edge.src,
                edge.destination
            )
        )
        self.storage[edge.src].add((edge.destination, edge.ppt))
        self.storage[edge.destination].add((edge.src, edge.ppt))

    def edges(self) -> set:
        return set(
            flatten(
                [[Edge(v, e[0], e[1]) if e[0] > v else Edge(e[0], v, e[1]) for e in edges] for v, edges in self.storage.items()]
            )
        )

    @staticmethod
    def from_json(json_str: str):
        data = json.loads(json_str)
        graph = AdjListUndirectedDiGraph()
        for v in data.keys():
            graph.add_vertex(v)

        for v, edges in data.items():
            for destination, ppt in edges:
                graph.add_edge(Edge(v, destination, ppt))

        return graph




from abc import ABC, abstractmethod


class Heap (ABC):
    def __init__(self):
        self.store = list()

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        return iter(self.store)

    @abstractmethod
    def _compare(self, left, right) -> bool:
        pass

    @staticmethod
    def _get_parent(idx: int):
        return (idx - 1) // 2

    @staticmethod
    def _get_left_child(idx: int):
        return idx * 2 + 1

    @staticmethod
    def _get_right_child(idx: int):
        return idx * 2 + 2

    def _get_swappable_child(self, idx: int):
        left = Heap._get_left_child(idx)
        right = Heap._get_right_child(idx)
        if right >= len(self.store):
            return left
        else:
            return right if self._compare(
                self.store[left].__key__(), self.store[right].__key__()
            ) else left

    def _swap(self, left: int, right: int):
        tmp = self.store[left]
        self.store[left] = self.store[right]
        self.store[right] = tmp

    def _heapify(self, idx: int):
        # Bubble up if needed
        curr_idx = idx
        parent = Heap._get_parent(idx)
        while curr_idx > 0 and not self._compare(self.store[curr_idx].__key__(), self.store[parent].__key__()):
            self._swap(curr_idx, parent)
            curr_idx = parent
            parent = Heap._get_parent(curr_idx)

        # Bubble down if needed
        child = self._get_swappable_child(curr_idx)
        while child < len(self.store) and self._compare(self.store[curr_idx].__key__(), self.store[child].__key__()):
            self._swap(curr_idx, child)
            curr_idx = child
            child = self._get_swappable_child(curr_idx)

    def insert(self, elem):
        """
        Insert an element into the heap. The element has to have an attribute called `__key__`  which provides the key
        for comparison.
        :param elem: The element to insert into the heap
        :return
        """
        if not hasattr(elem, "__key__"):
            raise TypeError("{} does not define the __key__ attribute that is needed for heap operations")

        self.store.append(elem)
        self._heapify(len(self.store) - 1)

    def extract_extreme(self):
        """
        Extract the top of the heap
        """
        if len(self) == 0:
            return None
        elem = self.store[0]
        self.store[0] = self.store[-1]
        del self.store[-1]
        if len(self) > 0:
            self._heapify(0)
        return elem

    def get_extreme(self):
        """
        Returns the top element in the heap without extracting it
        """
        return self.store[0]

    def build_heap(self):
        """
        This re-builds the full heap. This is necessary if sometime there are many key updates for the heap elements
        externally and it is necessary to rebuild the whole heap.
        """
        for i in range(len(self) // 2, -1, -1):
            self._heapify(i)


class MinHeap (Heap):
    """
    Implements the min-heap data structure, where the element with the smallest key is always at the top if the heap.
    """
    def _compare(self, left, right) -> bool:
        return left >= right

    @staticmethod
    def from_iterable(collection):
        heap = MinHeap()
        for elem in collection:
            heap.insert(elem)
        return heap


class MaxHeap (Heap):
    """
    Implements the max-heap data structure, where the element with the largest key is always at the top if the heap.
    """
    def _compare(self, left, right) -> bool:
        return left <= right

    @staticmethod
    def from_iterable(collection):
        heap = MaxHeap()
        for elem in collection:
            heap.insert(elem)
        return heap


from thesoup.utilityclasses.graph import Graph
from thesoup.utilityclasses.heap import MinHeap
from thesoup.utilityfunctions.collectionutils import flatten, flatten_to_tuple


def bfs(graph: Graph, start) -> dict:
    """
    This function implements bread first search on an object of type Graph. Complexity of such an algorithm is upper
    bound by `O(v)` where `v` is the number of vertices.

    It returns a map representing the predecessor relationship and "levels" of vertices. Example, consider the graph

    ```
    A -- B
     `-- C-- D
          `-- E
    ```

    Starting from A, the `bfs` function will return a map like this:
    {
      "A": (None, 0),
      "B": ("A", 1),
      "C": ("A", 1),
      "D": ("C", 2),
      "E": ("C",2)
    }
    This structure can be used to trace paths between the start to any reachable vertex.

    :param graph: The Digraph to traverse
    :param start: The starting point
    :return: A map of predecessors and levels.
    """
    predecessors = dict()
    if start in graph:
        predecessors[start] = (None, 0)
    else:
        return dict()
    frontier = {start}
    level = 1
    while len(frontier) > 0:
        next_frontier_unfiltered = flatten_to_tuple(
                [set(map(
                    lambda node: (node[0], (u, level)),
                    graph.get_neighbours(u)
                )) for u in frontier]
            )
        next_frontier = set(filter(
            lambda elem: elem[0] not in predecessors,
            next_frontier_unfiltered
        ))
        predecessors.update(next_frontier)
        frontier = set(
            map(
                lambda item: item[0],
                next_frontier
            )
        )
        level += 1
    return predecessors


def _dfs_callback(graph: Graph, start, parents: dict):
    for n in map(lambda x: x[0], graph.get_neighbours(start)):
        if n not in parents:
            parents[n] = start
            _dfs_callback(graph, n, parents)


def dfs(graph: Graph, start):
    """
    This implements depth first search on an object of type Graph. Complexity of such an algorithm is upper bound by
    `O(v)` where `v` is the number of vertices.

    It returns a map representing the predecessor relationship. Example consider the graph

    ```
    A -- B
     `-- C-- D
          `-- E
    ```

    Starting from A, the `dfs` function will return a map like this:
    {
      "A": None,
      "B": "A",
      "C": "A",
      "D": "C",
      "E": "C"
    }
    This structure can be used to trace paths between the start to any reachable vertex.


    :param graph: The Digraph to traverse
    :param start: The starting point
    :return: A map of traversed vertices vs it's parents.
    """
    if start not in graph:
        return dict()
    parents = {start: None}
    _dfs_callback(graph, start, parents)
    return parents


def dijkstra(graph: Graph, start) -> (dict, dict):
    """
    This implements the dijkstra's algorithm fpr shortest path. It returns a tuple containing 2 dictionaries:
    a map of distances of vertices from `start` and another congaing the predecessor of each vertex.

    NOTE: The input type is `Graph`. There is no separate interface for a di-graph in this library. Implementations of
    the `Graph` class must make it a di-graph. Otherwise this algorithm will hang.

    :param graph: The Digraph to traverse
    :param start: The starting point
    :return: A tuple (distances from start, predecessors).
    """
    if start not in graph:
        return dict(), dict()

    # Init
    d = dict([(v, float('inf')) for v in graph.vertices()])
    predecessors = dict([(v, None) for v in graph.vertices()])
    d[start] = 0

    class HeapElement:
        def __init__(self, vertex):
            self.vertex = vertex

        def __key__(self):
            return d[self.vertex]

    heap = MinHeap.from_iterable(list(map(lambda v: HeapElement(v), d.keys())))

    # Relax and repeat
    while len(heap) > 0:
        min_elem = heap.extract_extreme()
        u = min_elem.vertex
        for v, ppt in graph.get_neighbours(u):
            if d[v] > d[u] + ppt:
                d[v] = d[u] + ppt
                predecessors[v] = u
        heap.build_heap()
    return d, predecessors


def _shortest_path_dag_explorer(graph: Graph, predecessors: dict, memo: dict, start, end):
    if start == end:
        return 0
    elif (start, end) in memo:
        return memo[(start, end)]
    else:
        neighbours = graph.get_neighbours(start)
        if neighbours is None:
            return float('inf')
        if len(neighbours) == 0:
            return float('inf')
        sp, n = min(
            [(_shortest_path_dag_explorer(graph, predecessors, memo, n, end) + ppt, n) for n, ppt in neighbours]
        )
        if sp != float('inf'):
            memo[(start, end)] = sp
            predecessors[n] = start
        return sp


def shortest_path_dag(graph: Graph, start, end) -> (float, dict):
    """
    This function implements a special case SP algorithm for DAGs. Since it's a DAG, we can do this in a dynamic graph
    as well.
    NOTE: This
    :param graph: The Digraph to traverse
    :param start: The starting point
    :param end: The ending point
    :return: A tuple (shortest path, predecessors map).
    """
    memo = dict()
    predecessors = {start: None}
    sp = _shortest_path_dag_explorer(graph, predecessors, memo, start, end)
    return sp, predecessors



from thesoup.utilityclasses.sets import CountSet


def is_anagram(s1: str, s2: str) -> bool:
    """
    This function takes 2 strings as input and tells if they are anagrams of each other. 2 strings are anagrams if they
    can be constructed by rearranging the other word.

    :param s1: The first string
    :param s2: The second string
    """
    counts1 = CountSet(s1)
    counts2 = CountSet(s2)
    return counts1 == counts2
